import pytest
from ke.parser import Concat, Either, Def, Operator, Macro, Range, Literal, Nothing
from ke import asm
from ke.compiler import compile as _compile, CompileError


def compile(ast):
    result = _compile(ast)
    assert result.setting == ""
    return result.sub


def test_literal():
    assert compile(Literal("abc")) == asm.Literal("abc")


def test_concat():
    assert compile(Concat(list(map(Literal, "abc")))) == asm.Concat(
        [asm.Literal("a"), asm.Literal("b"), asm.Literal("c")]
    )


def test_either():
    assert compile(Either([Literal("abc"), Literal("def")])) == asm.Either(
        [asm.Literal("abc"), asm.Literal("def")]
    )


def test_macro():
    assert compile(Macro("#digit")) == asm.DIGIT


def test_nothing():
    assert compile(Nothing()) == asm.Literal("")
    assert compile(Concat([Nothing(), Nothing()])) == asm.Literal("")


def test_op():
    assert compile(Operator("capture", None, Literal("Yo"))) == asm.Capture(
        None, asm.Literal("Yo")
    )
    assert compile(Operator("capture", "name", Literal("Yo"))) == asm.Capture(
        "name", asm.Literal("Yo")
    )
    assert compile(Operator("0-1", None, Literal("Yo"))) == asm.Multiple(
        0, 1, True, asm.Literal("Yo")
    )
    assert compile(Operator("1+", None, Literal("Yo"))) == asm.Multiple(
        1, None, True, asm.Literal("Yo")
    )


def test_def():
    assert compile(Concat([Def("#x", Literal("x")), Macro("#x")])) == asm.Literal("x")
    assert compile(Concat([Macro("#x"), Def("#x", Literal("x"))])) == asm.Literal("x")


def test_def_scoping():
    assert compile(
        Concat([Concat([Macro("#x")]), Def("#x", Literal("x"))])
    ) == asm.Literal("x")
    with pytest.raises(CompileError):
        compile(Concat([Concat([Def("#x", Literal("x"))]), Macro("#x")]))


def test_builtin_macros():
    assert compile(Macro("#any")) == asm.ANY
    not_linefeed = asm.CharacterClass([r"\n"], inverted=True)
    assert compile(Macro("#not_linefeed")) == not_linefeed
    assert compile(Macro("#windows_newline")) == asm.Literal("\r\n")
    assert compile(Concat([Macro("#sl"), Literal("yo"), Macro("#el")])) == asm.Concat(
        [asm.Boundary("^", None), asm.Literal("yo"), asm.Boundary("$", None)]
    )
    assert compile(Macro("#quote")) == asm.Literal("'")
    assert compile(Macro("#double_quote")) == asm.Literal('"')
    assert compile(Macro("#left_brace")) == asm.Literal("[")
    assert compile(Macro("#right_brace")) == asm.Literal("]")


def test_short_names():
    macro_names = [
        ("linefeed", "lf"),
        ("not_linefeed", "nlf"),
        ("carriage_return", "cr"),
        ("not_carriage_return", "ncr"),
        ("tab", "t"),
        ("not_tab", "nt"),
        ("digit", "d"),
        ("not_digit", "nd"),
        ("letter", "l"),
        ("not_letter", "nl"),
        ("lowercase", "lc"),
        ("not_lowercase", "nlc"),
        ("uppercase", "uc"),
        ("not_uppercase", "nuc"),
        ("space", "s"),
        ("not_space", "ns"),
        ("token_character", "tc"),
        ("not_token_character", "ntc"),
        ("word_boundary", "wb"),
        ("not_word_boundary", "nwb"),
        ("any", "a"),
        ("windows_newline", "crlf"),
        ("start_string", "ss"),
        ("end_string", "es"),
        ("start_line", "sl"),
        ("end_line", "el"),
        ("quote", "q"),
        ("double_quote", "dq"),
        ("left_brace", "lb"),
        ("right_brace", "rb"),
        ("integer", "int"),
        ("unsigned_integer", "uint"),
        ("hex_digit", "hexd"),
        ("hex_number", "hexn"),
    ]
    for long, short in macro_names:
        assert compile(Macro("#" + long)) == compile(Macro("#" + short))


def test_range_macros():
    assert compile(Range("a", "f")) == asm.CharacterClass([("a", "f")], False)
    assert compile(Range("B", "Z")) == asm.CharacterClass([("B", "Z")], False)
    assert compile(Range("2", "6")) == asm.CharacterClass([("2", "6")], False)
    assert compile(Either([Range("a", "f"), Macro("#digit")])) == asm.CharacterClass(
        [("a", "f"), r"\d"], False
    )
    with pytest.raises(CompileError):
        compile(Range("a", "5"))
    with pytest.raises(CompileError):
        compile(Range("a", "F"))
    with pytest.raises(CompileError):
        compile(Range("c", "a"))
    with pytest.raises(CompileError):
        compile(Range("a", "a"))
    with pytest.raises(AssertionError):
        compile(Range("!", ","))


def test_character_class():
    assert compile(Either([])) == asm.CharacterClass([], inverted=False)
    assert compile(Either([Literal("a"), Literal("b")])) == asm.CharacterClass(
        ["a", "b"], inverted=False
    )
    assert compile(
        Operator("not", None, Either([Literal("a"), Literal("b")]))
    ) == asm.CharacterClass(["a", "b"], inverted=True)
    with pytest.raises(CompileError):
        compile(Operator("not", None, Either([Literal("a"), Literal("bc")])))
    assert compile(
        Either([Literal("a"), Literal("b"), Literal("0")])
    ) == asm.CharacterClass(["a", "b", "0"], inverted=False)
    assert compile(Either([Literal("a"), Macro("#d")])) == asm.CharacterClass(
        ["a", r"\d"], inverted=False
    )
    assert compile(Either([Literal("["), Literal("]")])) == asm.CharacterClass(
        ["[", "]"], inverted=False
    )
    assert compile(
        Concat([Either([Literal("["), Literal("]")])])
    ) == asm.CharacterClass(["[", "]"], inverted=False)


def test_invert():
    assert compile(Operator("not", None, Either([]))) == asm.CharacterClass(
        [], inverted=True
    )
    assert compile(Operator("not", None, Literal("a"))) == asm.CharacterClass(
        ["a"], inverted=True
    )
    assert compile(
        Operator("not", None, Either([Literal("a"), Literal("b")]))
    ) == asm.CharacterClass(["a", "b"], inverted=True)
    assert compile(
        Operator("not", None, Either([Literal("a"), Macro("#d")]))
    ) == asm.CharacterClass(["a", r"\d"], inverted=True)
    assert compile(
        Operator("not", None, Either([Literal("a"), Macro("#l")]))
    ) == asm.CharacterClass(["a", ["a", "z"], ["A", "Z"]], inverted=True)


def test_comment():
    assert compile(Operator("comment", None, Either([]))) == asm.Literal("")
    assert compile(Operator("comment", None, Literal("a"))) == asm.Literal("")
    assert compile(
        Operator("comment", None, Either([Literal("a"), Literal("b")]))
    ) == asm.Literal("")
    assert compile(
        Concat([Macro("#sl"), Operator("comment", None, Literal("yo")), Macro("#el")])
    ) == asm.Concat([asm.Boundary("^", None), asm.Boundary("$", None)])



import re
import parsimonious


class KleenexpError(re.error):
    pass


error = KleenexpError
# we want to raise the nice parsimonious ParseError with all the explanation,
# but we also want to raise something that isinstance(x, re.error)...
class ParseError(parsimonious.ParseError, KleenexpError):
    pass


class CompileError(KleenexpError):
    pass



import re
from collections import namedtuple

PYTHON_IDENTIFIER = re.compile("^[a-z_][a-z0-9_]*$", re.I)


class Asm(object):
    def to_regex(self, syntax, wrap=False):
        raise NotImplementedError()

    def maybe_wrap(self, should_wrap, regex):
        if not should_wrap:
            return regex
        return "(?:%s)" % regex


# python's standard re.escape() with our adjustments
_special_chars_map = {i: "\\" + chr(i) for i in b"()[]{}?*+|^$\\. \t\n\r\v\f"}

_kleen_special_chars_map = {
    ord("\t"): "t",
    ord("\n"): "n",
    ord("\r"): "r",
    ord("\v"): "v",
    ord("\f"): "f",
}

_char_escapes = {chr(k): "\\" + v for k, v in _kleen_special_chars_map.items()}


class Literal(namedtuple("Literal", ["string"]), Asm):
    def to_regex(self, syntax, wrap=False):
        escaped = (
            self.escape(self.string)
            .translate(_kleen_special_chars_map)
            .replace("\\ ", " ")
        )
        return self.maybe_wrap(wrap and len(self.string) != 1, escaped)

    def escape(self, pattern):
        """
        Escape special characters in a string.
        """
        if isinstance(pattern, str):
            return pattern.translate(_special_chars_map)
        else:
            pattern = str(pattern, "latin1")
            return pattern.translate(_special_chars_map).encode("latin1")


class Multiple(namedtuple("Multiple", ["min", "max", "is_greedy", "sub"]), Asm):
    def to_regex(self, syntax, wrap=False):
        if self.min == 0 and self.max is None:
            op = "*"
        elif self.min == 1 and self.max is None:
            op = "+"
        elif self.min == 0 and self.max == 1:
            op = "?"
        elif self.min == self.max:
            op = "{%d}" % self.min
        else:
            op = "{%s,%s}" % (self.min or "", self.max or "")
        if not self.is_greedy:
            op += "?"
        return self.maybe_wrap(wrap, self.sub.to_regex(syntax, wrap=True) + op)


class Either(namedtuple("Either", ["subs"]), Asm):
    def to_regex(self, syntax, wrap=False):
        return self.maybe_wrap(
            wrap, "|".join(s.to_regex(syntax, wrap=False) for s in self.subs)
        )


class Concat(namedtuple("Concat", ["subs"]), Asm):
    def to_regex(self, syntax, wrap=False):
        return self.maybe_wrap(
            wrap,
            "".join(s.to_regex(syntax, wrap=self.should_wrap(s)) for s in self.subs),
        )

    def should_wrap(self, s):
        return isinstance(s, Either) and len(self.subs) > 1


class CharacterClass(namedtuple("CharacterClass", ["characters", "inverted"]), Asm):
    def to_regex(self, syntax, wrap=False):
        if len(self.characters) == 0:
            if self.inverted:
                return "."  # Requires DOTALL flag.
            else:
                # Empty class (how did we get one?!) => need an expression that never matches.
                # http://stackoverflow.com/a/942122/239657: a lookahead of empty string
                # always matches, negative lookahead never does.
                # The dot afterward is to express an exactly-one-char pattern,
                # though it won't matter.
                return "(?!)."
        if len(self.characters) == 1:
            c = self.characters[0]
            if isinstance(c, str):
                if not self.inverted:
                    if len(c) == 1:
                        if c in _char_escapes:
                            return _char_escapes[c]
                        return Literal(c).to_regex(syntax)
                    return c
                elif c.startswith("\\") and c[1:] in ("d", "s", "w"):
                    return c.upper()
        return "[%s%s]" % ("^" if self.inverted else "", self.join_characters())

    NEED_ESCAPING = "^-[]\\"

    def escape_char(self, c):
        return _char_escapes.get(c, "\\" + c if c in self.NEED_ESCAPING else c)

    def join_characters(self):
        return "".join(
            sorted(
                self.escape_char(c)
                if isinstance(c, str)
                else "-".join(map(self.escape_char, c))
                for c in self.characters
            )
        )

    def invert(self):
        return CharacterClass(self.characters, not self.inverted)


ANY = CharacterClass([], inverted=True)
NEWLINE = CharacterClass([r"\r", r"\n", r"\u2028", r"\u2029"], False)
LINEFEED = CharacterClass([r"\n"], False)
CARRIAGE_RETURN = CharacterClass([r"\r"], False)
TAB = CharacterClass([r"\t"], False)
DIGIT = CharacterClass([r"\d"], False)
LETTER = CharacterClass([["a", "z"], ["A", "Z"]], False)
LOWERCASE = CharacterClass([["a", "z"]], False)
UPPERCASE = CharacterClass([["A", "Z"]], False)
SPACE = CharacterClass([r"\s"], False)
TOKEN_CHARACTER = CharacterClass([r"\w"], False)


class Boundary(namedtuple("Boundary", ["character", "reverse"]), Asm):
    def to_regex(self, syntax, wrap=False):
        return self.character

    def invert(self):
        if self.reverse is None:
            raise ValueError("Cannot invert %s" % (self,))
        return Boundary(self.reverse, self.character)


START_LINE = Boundary(r"^", None)
END_LINE = Boundary(r"$", None)
START_STRING = Boundary(r"\A", None)
# TODO: \z matches purely at string end; \Z also matches before \n\z.
# Should we expose \z or the pragmatic \Z you usually want?  Or both?
END_STRING = Boundary(r"\Z", None)
WORD_BOUNDARY = Boundary(r"\b", r"\B")


class Capture(namedtuple("Capture", ["name", "sub"]), Asm):
    def to_regex(self, syntax, wrap=False):
        return "(%s%s)" % (
            self.name_regex(syntax),
            self.sub.to_regex(syntax, wrap=False),
        )

    def name_regex(self, syntax):
        if self.name is None:
            return ""
        if not self.name:
            raise ValueError("Capture name cannot be empty")
        if not PYTHON_IDENTIFIER.match(self.name):
            raise ValueError("invalid capture group name: %s" % self.name)
        if syntax == "javascript":
            return "?<%s>" % self.name
        return "?P<%s>" % self.name


class ParensSyntax(namedtuple("ParensSyntax", ["name", "sub"]), Asm):
    INVERTED = None

    def to_regex(self, syntax, wrap=False):
        if self.name is not None:
            self.error("doesn't support naming")
        return f"({self.HEADER}{self.sub.to_regex(syntax, wrap=False)})"

    def invert(self):
        if self.INVERTED is None:
            self.error("doesn't support inverting")
        return self.INVERTED(None, self.sub)

    def error(self, message):
        raise ValueError(f"{self.__class__.__name__} {message}")


class Lookahead(ParensSyntax):
    HEADER = "?="


class NegativeLookahead(ParensSyntax):
    HEADER = "?!"
    INVERTED = Lookahead


Lookahead.INVERTED = NegativeLookahead


class Lookbehind(ParensSyntax):
    HEADER = "?<="


class NegativeLookbehind(ParensSyntax):
    HEADER = "?<!"
    INVERTED = Lookbehind


Lookbehind.INVERTED = NegativeLookbehind


class Setting(namedtuple("Setting", ["setting", "sub"]), Asm):
    def to_regex(self, syntax, wrap=False):
        if not self.setting:
            return self.sub.to_regex(syntax, wrap=False)
        # no need to wrap because settings have global effect and match 0 characters,
        # so e.g. /(?m)ab|c/ == /(?:(?m)ab)|c/, however, child may need to wrap
        # or we risk accidental /(?m)ab?/ instead of /(?m)(ab)?/
        return "(?%s)%s" % (self.setting, self.sub.to_regex(syntax, wrap=wrap))


def assemble(asm, syntax="python"):
    return asm.to_regex(syntax, wrap=False)



import pytest
import midas
import ke
import re


def assert_pattern(pattern, matches, not_matches=()):
    if isinstance(pattern, str):
        pattern = ke.compile(pattern)
    for m in matches:
        assert pattern.match(m), (pattern, m)
    for m in not_matches:
        assert not pattern.match(m), (pattern, m)


def test_assumptions():
    assert_pattern(re.compile(r".*a.c"), ["abc", "ab abc"], ["", "ac", "\nabc"])
    assert_pattern(re.compile(r"(?:.|[\r\n])*a.c"), ["\nabc", "\n\n\nabc"], ["ab"])


def test_api_equivalence():
    ke_dir = set(x for x in dir(ke) if not x.startswith("_"))
    re_dir = set(re.__all__)
    re_dir.difference_update(["template"])  # experimental, not documented
    assert re_dir
    assert re_dir - ke_dir == set()


@midas.test(format="lines")
def test_compile_gold(ke_pattern):
    return ke.re(ke_pattern)


JABBERWOCKY = """\
'Twas brillig, and the slithy toves
Did gyre and gimble in the wabe:
All mimsy were the borogoves,
And the mome raths outgrabe.

"Beware the Jabberwock, my son!
The jaws that bite, the claws that catch!
Beware the Jubjub bird, and shun
The frumious Bandersnatch!"

He took his vorpal sword in hand:
Long time the manxome foe he sought,
So rested he by the Tumtum tree,
And stood a while in thought.

And, as in uffish thought he stood,
The Jabberwock, with eyes of flame,
Came whiffling through the tulgey wood,
And burbled as it came!

One two! One two! And through and through
The vorpal blade went snicker-snack!
He left it dead, and with its head
He went galumphing back.

"And hast thou slain the Jabberwock?
Come to my arms, my beamish boy!
Oh frabjous day! Callooh! Callay!"
He chortled in his joy.

'Twas brillig, and the slithy toves
Did gyre and gimble in the wabe:
All mimsy were the borogoves,
And the mome raths outgrabe.
"""


@midas.test(format="lines")
def test_jabberwocky(ke_pattern):
    return ke.findall(ke_pattern, JABBERWOCKY, ke.MULTILINE | ke.DOTALL | ke.IGNORECASE)


def test_re():
    assert ke.compile("\t\r\n").match("\t\r\n")


def test_redundant_backslash_escape():
    assert ke.re("-") == "-"
    assert ke.re("#") == "#"


def test_compile():
    assert ke.compile('["a"]').search("bab")
    assert not ke.compile('["c"]').search("bab")
    assert ke.compile('["a"]').search("bab")


def test_compile_bytes():
    assert ke.compile(b'["a"]').search(b"bab")
    assert not ke.compile(b'["c"]').search(b"bab")


def test_flags():
    assert ke.compile("a", ke.I).match("A")
    assert ke.compile("a[#el]", ke.M).search("a\nb")
    assert not ke.compile("a[#el]").search("a\nb")

    for flag in "AIUMS":
        k = getattr(ke, flag)
        r = getattr(re, flag)
        if r != re.ASCII:
            r |= re.UNICODE
        assert ke.compile("a", k).flags == r
    assert ke.compile(b"a", ke.LOCALE).flags == re.LOCALE
    assert ke.compile("a", ke.DEBUG).flags == re.DEBUG | re.UNICODE

    # X should have no effect
    assert ke.compile("a", ke.X).flags == re.UNICODE
    assert ke.re("[c #wb [1-3 #d] #wb]", ke.X) == ke.re("[c #wb [1-3 #d] #wb]")


def test_search():
    assert ke.search('["a"]', "xabc")
    assert not ke.search('["a"]', "Abc")
    assert ke.search('["a"]', "xabc", flags=ke.I)


def test_match():
    assert ke.match('["a"]', "abc")
    assert not ke.match('["a"]', "bac")
    assert ke.match('[capture "a"][capture:g "b"]', "abc").group(1) == "a"
    assert ke.match('[capture "a"][capture:g "b"]', "abc").group("g") == "b"
    assert not ke.match("a", "Ac")
    assert ke.match("a", "Ac", ke.I)


def test_fullmatch():
    assert ke.fullmatch('["a"]', "a")
    assert not ke.fullmatch('["a"]', "abc")


def test_split():
    assert ke.split("[#d]", "a1a2a3a4a5", maxsplit=3) == ["a", "a", "a", "a4a5"]


def test_findall():
    assert ke.findall("a", "xabca") == re.findall("a", "xabca")
    assert ke.findall("aaa", "aaaabaaab") == re.findall("aaa", "aaaabaaab")
    assert ke.findall("aaa", "aaAabaaab", ke.I) == re.findall("aaa", "aaAabaaab", re.I)


def test_finditer():
    assert [m.groups() for m in ke.finditer("[c #letter]", "xa.ca")] == [
        m.groups() for m in re.finditer(r"(\w)", "xa.ca")
    ]


def test_sub():
    assert (
        ke.sub(
            "Hi [capture 1+ #letter], what's up?",
            r"\1! \1!",
            "Hi Bobby, what's up? Hi Martin, what's up?",
        )
        == "Bobby! Bobby! Martin! Martin!"
    )
    assert ke.sub("[1+ #d]", "###", "123-45-6789", count=2) == "###-###-6789"
    assert (
        ke.sub("[c 1+ #d]", lambda m: m.group(1)[::-1], "123-45-6789") == "321-54-9876"
    )
    with pytest.raises(IndexError):
        ke.sub("[1+ #d]", lambda m: m.group(1)[::-1], "123-45-6789")


def test_subn():
    assert ke.subn("[1+ #d]", "###", "123-45-6789", count=2) == ("###-###-6789", 2)


@midas.test(format="lines")
def test_escape(line):
    k = ke.escape(line)
    assert ke.match(k, line)
    assert ke.re(k) == re.escape(line)
    return k


def test_multiple():
    assert ke.match('[1+ "a"]', "a")
    assert ke.match('[1+ "a"]', "aa")
    assert not ke.match('[1+ "a"]', "")
    assert ke.match('[2-3 "a"][#end_line]', "aa")
    assert ke.match('[2-3 "a"][#end_line]', "aaa")
    assert not ke.match('[2-3 "a"][#end_line]', "a")
    assert not ke.match('[2-3 "a"][#end_line]', "aaaa")
    assert ke.match('[0-1 "a"][#end_line]', "")
    assert ke.match('[0-1 "a"][#end_line]', "a")
    assert not ke.match('[0-1 "a"][#end_line]', "aa")
    assert ke.match('[2 "a"][#end_line]', "aa")
    assert not ke.match('[2 "a"][#end_line]', "a")
    assert not ke.match('[2 "a"][#end_line]', "aaa")
    assert ke.match('[2-3 ["hi" | "bye"]][#end_line]', "byebye")
    assert ke.match('[2-3 ["hi" | "bye"]][#end_line]', "hibye")
    assert ke.match('[2-3 ["hi" | "bye"]][#end_line]', "hihibye")
    assert not ke.match('[2-3 ["hi" | "bye"]][#end_line]', "hihihihi")
    assert not ke.match('[2-3 ["hi" | "bye"]][#end_line]', "hi")


def test_capture():
    assert ke.compile('[capture 3-5 "a"]').match("aaa").group(1) == "aaa"
    assert ke.compile('[capture 3-5 "a"]').match("aaaaaaaaa").group(1) == "aaaaa"
    assert not ke.compile('[capture 3-5 "a"]').match("aa")
    assert not ke.compile('[capture 3-5 "a"][#end_line]').match("aaaaaa")
    assert not ke.compile('[[capture 3-5 "a"] #end_line]').match("aaaaaa")
    assert ke.match('[capture 3-5 "a"]', "aaaaaaaaa").group(1) == "aaaaa"

    with pytest.raises(re.error):
        ke.re("[capture 3-5]")
    with pytest.raises(re.error):
        ke.re("[capture 3-5 []]")
    with pytest.raises(re.error):
        ke.re('[capture 0 "a"]')


def test_named_capture():
    assert ke.compile('[capture:a 3-5 "a"]').match("aaa").group("a") == "aaa"


def test_comments():
    assert ke.re("[comment]") == ke.re("[]")
    assert ke.re('[comment "a"]') == ke.re("[]")
    assert ke.re("[comment #token]") == ke.re("[]")
    assert ke.re("[comment not #token]") == ke.re("[]")
    with pytest.raises(re.error):
        ke.re("[0-1 comment #token]")
    assert ke.re('["a" [comment "a"] "b"]') == ke.re("ab")
    assert ke.re('[comment [["a"]]]') == ke.re("[]")


def test_range_macros():
    assert ke.re("[#a..z]") == "[a-z]"
    assert ke.re('[#a..c | "g" | #q..t]') == "[a-cgq-t]"
    assert ke.re('[#a..c | "-"]') == r"[\-a-c]"
    assert ke.match('[#a..c | "-"]', "a")
    assert ke.match('[#a..c | "-"]', "b")
    assert ke.match('[#a..c | "-"]', "c")
    assert ke.match('[#a..c | "-"]', "-")
    assert not ke.match('[#a..c | "-"]', "d")
    with pytest.raises(re.error):
        ke.re("[#..]")
    with pytest.raises(re.error):
        ke.re("[#a..]")
    with pytest.raises(re.error):
        ke.re("[#a..a]")
    with pytest.raises(re.error):
        ke.re("[#a..em..z]")
    with pytest.raises(re.error):
        ke.re("[#!../ #:..@ #....]")


def test_not():
    assert ke.compile('[not "a"]').match("b")
    assert ke.compile('[not "a"]').match("A")
    assert not ke.compile('[not "a"]').match("a")

    assert not ke.compile('[not not "a"]').match("b")
    assert not ke.compile('[not not "a"]').match("A")
    assert ke.compile('[not not "a"]').match("a")

    assert ke.compile('[not ["a" | "b"]]').match("c")
    assert not ke.compile('[not ["a" | "b"]]').match("a")
    assert not ke.compile('[not ["a" | "b"]]').match("b")

    assert ke.compile('[not "a"]').match("0")
    assert ke.compile('[not ["a" | #d]]').match("b")
    assert not ke.compile('[not ["a" | #d]]').match("0")
    assert not ke.compile('[not ["a" | #d]]').match("a")
    assert not ke.compile('[not ["a" | #d]]').match("9")

    assert ke.compile("[not #a..f]").match("g")
    assert ke.compile("[not #a..f]").match("A")
    assert not ke.compile("[not #a..f]").match("a")
    assert not ke.compile("[not #a..f]").match("c")
    assert not ke.compile("[not #a..f]").match("f")

    with pytest.raises(re.error):
        ke.compile('[not "ab"]')
    with pytest.raises(re.error):
        ke.compile("[not]")


def test_lookahead():
    assert ke.compile('[lookahead "a"]a').match("a")
    assert not ke.compile('[lookahead "a"]b').match("a")
    assert not ke.compile('[lookahead "a"]b').match("b")
    assert ke.compile("[lookahead 5 #digit]1").match("12345")
    assert not ke.compile("[lookahead 5 #digit]1").match("1234")
    assert not ke.compile("[lookahead 5 #digit]1").match("22345")
    assert ke.compile("[lookahead [0+ #any] [not #digit]]1").match("1234a")
    assert not ke.compile("[lookahead [0+ #any] [not #digit]]1").match("12345")
    assert ke.compile('[not lookahead "ab"]a').match("a")
    assert ke.compile('[not lookahead "ab"]a').match("ac")
    assert not ke.compile('[not lookahead "ab"]a').match("ab")
    assert ke.compile("[not lookahead [0+ #any] [not #digit]]1").match("12345")
    assert not ke.compile("[not lookahead [0+ #any] [not #digit]]1").match("1234a")
    password_ke = ke.compile(
        """[#start_string
        [lookahead [0+ not #lowercase] #lowercase]
        [lookahead [0+ not #uppercase] #uppercase]
        [lookahead [0+ not #digit] [capture #digit]]
        [not lookahead [0+ #any] ["123" | "pass" | "Pass"]]
        [6+ #token_character]
        #end_string
    ]"""
    )
    print(password_ke.pattern)
    assert password_ke.match("pAssw1")
    assert password_ke.match("pAssword1")
    assert password_ke.match("pAssword1").group(1) == "1"
    assert not password_ke.match("PAss1")
    assert not password_ke.match("p4ssword1")
    assert not password_ke.match("PASSWORD1")
    assert not password_ke.match("PAssword")
    assert not password_ke.match("PAssword1#")


def test_lookbehind():
    assert ke.compile("[#letter][not lookbehind 'ab']").match("ac")
    assert ke.compile("[#letter][not lookbehind 'ab']").match("b")
    assert ke.compile("([[1+ #any] [lookbehind 3 #lowercase]])").search(
        "hi (and bye) dude"
    )
    assert not ke.compile("([[1+ #any] [lookbehind 3 #lowercase]])").search(
        "hi (and Bye) dude"
    )


def test_real():
    print(ke.re("[#ss #real #es]"))
    r = ke.compile("[#ss #real #es]")
    assert r.match("0")
    assert r.match("0.0")
    assert r.match("-0.0")
    assert r.match("1234.56")
    assert r.match("-0.0")
    assert not r.match("0.")
    assert not r.match(".0")
    assert not r.match("-0.")
    assert not r.match("-.0")


def test_float():
    print(ke.re("[#ss #float #es]"))
    f = ke.compile("[#ss #float #es]")
    assert f.match("0.0")
    assert f.match("-0.0")
    assert f.match("0.0e1")
    assert f.match("-0.0e1")
    assert f.match("0.0e-1")
    assert f.match("0.0E1")
    assert f.match("0.")
    assert f.match("0.e1")
    assert f.match(".0")
    assert f.match(".0e1")
    assert f.match("0e1")
    assert not f.match("0")
    assert not f.match(".")
    assert not f.match(".e1")
    assert not f.match("0.0e")
    assert f.match("1024.12e3")
    assert f.match("-1024.12e-3")
    assert f.match("-.12e3")
    assert f.match("-1024.12E-3")


def test_hex():
    h = ke.compile("[#ss #hexn #es]")
    assert h.match("0")
    assert h.match("9")
    assert h.match("a")
    assert h.match("f")
    assert h.match("A")
    assert h.match("1234567890abcdef")
    assert h.match("09af09AF")
    assert not h.match("-1")
    assert not h.match("g")


def test_token():
    h = ke.compile("[#ss #token #es]")
    assert h.match("a")
    assert h.match("abc")
    assert h.match("a1")
    assert h.match("A")
    assert h.match("AbC19")
    assert h.match("_")
    assert h.match("_a")
    assert h.match("_1")
    assert h.match("a_b_c")
    assert not h.match("1")
    assert not h.match("1_")
    assert not h.match("9_")
    assert not h.match("1234")
    assert not h.match("!")
    assert not h.match("a!")
    assert not h.match("#x")
    assert not h.match("x ")
    assert not h.match("x y")


def test_c0_c1():
    assert ke.compile("a[#c0]z").match("a16z").groups()[0] == "16"
    assert ke.compile("a[#c0]z").search("http://abc.xyz/").groups()[0] == "bc.xy"
    assert ke.compile("a[#c0]z").search("azure").groups()[0] == ""

    assert ke.compile("a[#c1]z").match("a16z").groups()[0] == "16"
    assert ke.compile("a[#c1]z").search("http://abc.xyz/").groups()[0] == "bc.xy"
    assert not ke.compile("a[#c1]z").search("azure")


def test_escapes():
    assert ke.compile(
        "[#dq #q #t #lb #rb #vertical_tab #formfeed #bell #backspace #el]"
    ).match(""""'\t[]\v\f\a\b""")


def test_define_macros():
    expected = "Yo dawg, I heard you like Yo dawg, I heard you like this, so I put some of this in your regex so you can recurse while you recurse, so I put some Yo dawg, I heard you like this, so I put some of this in your regex so you can recurse while you recurse in your Yo dawg, I heard you like this, so I put some of this in your regex so you can recurse while you recurse so you can recurse while you recurse".replace(
        ",", re.escape(",")
    )  # `,` in CPython, `\,` in PyPy
    assert (
        ke.re(
            """[#recursive_dawg][
        #yo=["Yo dawg, I heard you like "]
        #so_i_put=[", so I put some "]
        #in_your=[" in your "]
        #so_you_can=[" so you can "]
        #while_you=[" while you "]
        #dawg=[#yo "this" #so_i_put "of this" #in_your "regex" #so_you_can "recurse" #while_you "recurse"]
        #recursive_dawg=[#yo #dawg #so_i_put #dawg #in_your #dawg #so_you_can "recurse" #while_you "recurse"]
    ]"""
        )
        == expected
    )


def test_newlines():
    assert_pattern(
        "[#sl]a[0+ #any]x[0+ #any]b[#el]", ["axb", "azzzxzzzb"], ["a\nx\nb", "\naxb"]
    )
    assert_pattern(
        ke.compile("[#sl]a[0+ #any]x[0+ #any]b[#el]", ke.DOTALL), ["a\nx\nb"], ["\naxb"]
    )
    assert_pattern(
        ke.compile("[#sl]a[0+ #aaa]x[0+ #aaa]b[#el]"), ["a\nx\nb"], ["\naxb"]
    )
    assert_pattern(
        ke.compile(
            "[0+ #any][#sl]a[0+ #any]x[0+ #any]b[#el]", ke.DOTALL | ke.MULTILINE
        ),
        ["\naxb", "ax\naxb", "\naxb\n", "ax\nazzzxzzzb"],
    )
    assert_pattern(
        ke.compile("[0+ #any][#sl]a[0+ #any]x[0+ #any]b[#el]", ke.MULTILINE),
        [],
        ["\naxb", "ax\naxb", "\naxb\n", "ax\nazzzxzzzb"],
    )
    assert_pattern(
        ke.compile("[0+ #aaa][#sl]a[0+ #aaa]x[0+ #aaa]b[#el]", ke.MULTILINE),
        ["\naxb", "ax\naxb", "\naxb\n", "ax\nazzzxzzzb"],
    )

    assert (
        ke.re("[#start_string][#newline][#end_string]")
        == r"\A(?:[\n\r\u2028\u2029]|\r\n)\Z"
    )
    assert_pattern(
        ke.compile("[#start_string][#newline][#end_string]"),
        ["\r", "\n", "\u2028", "\u2029", "\r\n"],
        ["\n\n", "\n\r", "\r\r"],
    )
    assert_pattern(
        ke.compile("[#start_string][#newline_character][#end_string]"),
        ["\r", "\n", "\u2028", "\u2029"],
        ["\r\n", "\n\n", "\n\r", "\r\r"],
    )
    assert_pattern(ke.compile("a[#not_newline]c"), ["abc"], ["a\rc", "a\nc", "abbc"])


def test_js():
    assert ke.re("[capture:hi 'hi']") == "(?P<hi>hi)"
    assert ke.re("[capture:hi 'hi']", syntax="javascript") == "(?<hi>hi)"



import pytest
from parsimonious.exceptions import ParseError
from ke.parser import (
    Parser,
    Concat as C,
    Either as E,
    Def as D,
    Operator,
    Macro as M,
    Range as R,
    Literal as L,
    Nothing as N,
)

O = lambda name, sub: Operator(name, None, sub)

v = Parser()


def test_outer_literal():
    assert v.parse("") == C([])
    assert v.parse("literal") == C([L("literal")])


def test_braces():
    assert v.parse("[]") == C([])
    assert v.parse("a[]b") == C([L("a"), L("b")])


def test_inner_literal():
    assert v.parse("['literal']") == C([L("literal")])
    assert v.parse("['']") == C([L("")])
    assert v.parse("""['"']""") == C([L('"')])
    assert v.parse("""["'"]""") == C([L("'")])


def test_multiple_inner_literals():
    assert v.parse("['11' '2']") == C([L("11"), L("2")])
    assert v.parse("[   '11' \t\n\r\n '2' ]") == C([L("11"), L("2")])
    assert v.parse("['1' '2' '3']") == C([L("1"), L("2"), L("3")])
    assert v.parse("""["1" '2' '3']""") == C([L("1"), L("2"), L("3")])
    assert v.parse("""["1' '2' '3"]""") == C([L("1' '2' '3")])


def test_macro():
    assert v.parse("[#a]") == C([M("#a")])
    assert v.parse("[#aloHa19]") == C([M("#aloHa19")])
    assert v.parse("[#a #b]") == C([M("#a"), M("#b")])
    assert v.parse("[ #a ]") == C([M("#a")])


def test_macro_special_chars():
    for o in range(128):
        char = chr(o)
        for name in char, "a" + char, char * 3:
            print(name)
            if char.isalnum() or char in "`~!@$%^&*()-_+,./;<>?{}\\":
                assert v.parse("[#%s]" % name) == C([M("#%s" % name)]), name
            elif char.isspace() and name != char:
                pass
            else:
                with pytest.raises(ParseError):
                    v.parse("[#%s]" % name)
    assert v.parse("[#a\t#b]") == C([M("#a"), M("#b")])
    assert v.parse("[#a\n#b]") == C([M("#a"), M("#b")])
    assert v.parse("[#a\r#b]") == C([M("#a"), M("#b")])


def test_macro_illegal_chars():
    for char in r"""[]#='"|""" + chr(0) + chr(0x1F) + chr(0x7F):
        for name in char, "a" + char, char * 3:
            print(name)
            with pytest.raises(ParseError):
                v.parse("[#%s]" % name)
    with pytest.raises(ParseError):
        v.parse("[#a\v#b]")


def test_range_macros():
    assert v.parse("[#a..b]") == C([R("a", "b")])
    assert v.parse("[#0..5]") == C([R("0", "5")])
    assert v.parse("[#a..9]") == C([R("a", "9")])
    assert v.parse("[#....]") == C([M("#....")])


def test_op():
    assert v.parse("[op #a]") == C([O("op", M("#a"))])
    assert v.parse("[op]") == C([O("op", N())])
    assert v.parse("[o p #a]") == C([O("o", O("p", M("#a")))])
    with pytest.raises(ParseError):
        v.parse("[#a op]")
    with pytest.raises(ParseError):
        v.parse("[op #a op]")
    assert v.parse("[!@$%^&*()\n<>?]") == C([O("!@$%^&*()", O("<>?", N()))])
    assert v.parse("[op:name #a]") == C([Operator("op", "name", M("#a"))])


def test_recursive_braces():
    assert v.parse("[[]]") == C([])
    assert v.parse("[a #d [b #e]]") == C([O("a", C([M("#d"), O("b", M("#e"))]))])
    assert v.parse("[a #d [b #e] [c #f]]") == C(
        [O("a", C([M("#d"), O("b", M("#e")), O("c", M("#f"))]))]
    )
    assert v.parse("[[1+ #d]'th']") == C([O("1+", M("#d")), L("th")])
    assert v.parse("[[][[]][]]") == C([N(), N(), N()])
    assert v.parse("[[]['a'['b']]'c'[]]") == C([N(), C([L("a"), L("b")]), L("c"), N()])
    with pytest.raises(ParseError):
        v.parse("[op [] op]")


def test_either():
    assert v.parse("[#a | #b]") == C([E([M("#a"), M("#b")])])
    assert v.parse("[#a|#b]") == C([E([M("#a"), M("#b")])])
    assert v.parse("[#a | #b | #c]") == C([E([M("#a"), M("#b"), M("#c")])])
    assert v.parse("[op [#a | #b]]") == C([O("op", E([M("#a"), M("#b")]))])
    assert v.parse("[op [#a #b | #c]]") == C(
        [O("op", E([C([M("#a"), M("#b")]), M("#c")]))]
    )
    with pytest.raises(ParseError):
        v.parse("[op #a | #b]")
    with pytest.raises(ParseError):
        v.parse("[op #a #b | #c]")
    with pytest.raises(ParseError):
        v.parse("[#a | op #b]")
    assert v.parse("[a #d [b #e] [c #f]]") == C(
        [O("a", C([M("#d"), O("b", M("#e")), O("c", M("#f"))]))]
    )
    with pytest.raises(ParseError):
        v.parse("[#a|]")
    with pytest.raises(ParseError):
        v.parse("[op #a|]")
    with pytest.raises(ParseError):
        v.parse("[op | #a]")
    assert v.parse("[#d|'a']") == C([E([M("#d"), L("a")])])
    assert v.parse("['['|']']") == C([E([L("["), L("]")])])


def test_def():
    assert v.parse("[#a=[#x]]") == C([D("#a", M("#x"))])
    assert v.parse("[#a #a=[#x #y]]") == C([M("#a"), D("#a", C([M("#x"), M("#y")]))])


def test_real():
    assert v.parse(
        "[#save_num] Reasons To Switch To Kleenexp, The [#save_num]th Made Me [case_insensitive ['Laugh' | 'Cry']][#save_num=[capture 1+ #digit]]"
    ) == C(
        [
            M("#save_num"),
            L(" Reasons To Switch To Kleenexp, The "),
            M("#save_num"),
            L("th Made Me "),
            O("case_insensitive", E([L("Laugh"), L("Cry")])),
            D("#save_num", O("capture", O("1+", M("#digit")))),
        ]
    )
    assert (
        v.parse(
            """[
        [capture 0-1 #proto] [capture #domain] '.' [capture #tld] [capture #path]
            #proto=['http' [0-1 's'] '://']
            #domain=[1+ [#digit | #lowercase | '.' | '-']]
            #tld=[2-6 [#lowercase | '.']]
            #path=['/' [0+ ['/' | #alphanum | '.' | '-']]]
    ]"""
        )
        == C(
            [
                O("capture", O("0-1", M("#proto"))),
                O("capture", M("#domain")),
                L("."),
                O("capture", M("#tld")),
                O("capture", M("#path")),
                D("#proto", C([L("http"), O("0-1", L("s")), L("://")])),
                D(
                    "#domain",
                    O("1+", E([M("#digit"), M("#lowercase"), L("."), L("-")])),
                ),
                D("#tld", O("2-6", E([M("#lowercase"), L(".")]))),
                D(
                    "#path",
                    C([L("/"), O("0+", E([L("/"), M("#alphanum"), L("."), L("-")]))]),
                ),
            ]
        )
    )



from parsimonious.grammar import Grammar
from parsimonious.nodes import NodeVisitor
from collections import namedtuple

grammar = Grammar(
    r"""
regex           = ( outer_literal / braces )*
braces          = '[' whitespace? ( ops_matches / either / matches )? whitespace? ']'
ops_matches     = op ( whitespace op )* ( whitespace matches )?
op              = token (':' token)?
either          = matches ( whitespace? '|' whitespace? matches )+
matches         = match ( whitespace match )*
match           = inner_literal / def / macro / braces
macro           = '#' ( range_macro / token )
range_macro     = range_endpoint '..' range_endpoint
def             = macro '=' braces

outer_literal   = ~r'[^\[\]]+'
inner_literal   = ( '\'' until_quote '\'' ) / ( '"' until_doublequote '"' )
until_quote     = ~r"[^']*"
until_doublequote = ~r'[^"]*'

# if separating between something and a brace, whitespace can be optional without introducing ambiguity
whitespace      = ~r'[ \t\r\n]+|(?<=\])|(?=\[)'
# '=' and ':' have syntactic meaning
token           = ~r'[A-Za-z0-9!$%&()*+,./;<>?@\\^_`{}~-]+'
range_endpoint  = ~r'[A-Za-z0-9]'
"""
)

Concat = namedtuple("Concat", ["items"])
Either = namedtuple("Either", ["items"])
Def = namedtuple("Def", ["name", "subregex"])
Operator = namedtuple("Operator", ["op_name", "name", "subregex"])
Macro = namedtuple("Macro", ["name"])
Range = namedtuple("Range", ["start", "end"])
Literal = namedtuple("Literal", ["string"])


class Nothing(object):
    def __eq__(self, other):
        return type(other) == type(self)


class Parser(NodeVisitor):
    grammar = grammar

    def generic_visit(self, node, visited_children):
        return visited_children or node

    def visit_regex(self, regex, nodes):
        flattened = []
        for (node,) in nodes:
            if isinstance(node, Concat):
                flattened += node.items
            elif not isinstance(node, Nothing):
                flattened.append(node)
        return Concat(flattened)

    def visit_braces(self, braces, data):
        (_l, _lw, in_braces, _rw, _r) = data
        in_braces = list(in_braces)
        if in_braces:
            (in_braces,), = in_braces
        else:
            in_braces = Nothing()
        assert type(in_braces) in [
            Concat,
            Either,
            Def,
            Operator,
            Literal,
            Macro,
            Range,
        ] or isinstance(in_braces, Nothing), in_braces
        return in_braces

    def visit_in_braces(self, in_braces, data):
        (ast,) = data
        return ast

    def visit_ops_matches(self, ops_matches, data):
        (op, more_ops, maybe_matches) = data
        ops = [op]
        for _w, op in more_ops:
            ops.append(op)

        maybe_matches = list(maybe_matches)
        if maybe_matches:
            (_w, result), = maybe_matches
        else:
            result = Nothing()
        while ops:
            op_name, name = ops.pop()
            name = list(name)
            if name:
                (_colon, name), = name
            else:
                name = None
            result = Operator(op_name, name, result)
        return result

    def visit_either(self, matches, data):
        (match, more_matches) = data
        more_matches = list(more_matches)
        assert more_matches
        result = [match]
        for _w1, _pipe, _w2, match in more_matches:
            result.append(match)
        return Either(result)

    def visit_matches(self, or_body, data):
        (match, more_matches) = data
        more_matches = list(more_matches)
        if not more_matches:
            return match
        result = [match]
        for _w, match in more_matches:
            result.append(match)
        return Concat(result)

    visit_match = NodeVisitor.lift_child

    def visit_macro(self, macro, data):
        (_hashtag, (parsed,)) = data
        if isinstance(parsed, Range):
            return parsed
        return Macro(macro.text)

    def visit_range_macro(self, range_macro, data):
        (start, _dotdot, end) = data
        return Range(start.text, end.text)

    def visit_def(self, _literal, data):
        (macro, _eq, braces) = data
        return Def(macro.name, braces)

    def visit_outer_literal(self, literal, _):
        return Literal(literal.text)

    def visit_inner_literal(self, _literal, data):
        ((_1, literal, _2),) = data
        return Literal(literal.text)

    def visit_token(self, token, _):
        return token.text



import functools
from typing import AnyStr, Callable
from parsimonious.exceptions import ParseError as ParsimoniousParseError
import argparse
import re as original_re
from re import (
    ASCII,
    A,
    IGNORECASE,
    I,
    LOCALE,
    L,
    UNICODE,
    U,
    MULTILINE,
    M,
    DOTALL,
    S,
    DEBUG,
    Match,
    Pattern,
)
import sys
import traceback

from ke.parser import Parser
from ke import compiler
from ke import asm
from ke.errors import KleenexpError, error, ParseError

VERBOSE = X = 0


ke_parser = Parser()


def _is_bytes_like(obj):
    return not hasattr(obj, "encode")


def re(pattern, syntax="python"):
    # TODO: LRU cache
    if _is_bytes_like(pattern):
        return re(pattern.decode("ascii")).encode("ascii")
    if syntax is None:
        syntax = "python"
    try:
        ast = ke_parser.parse(pattern)
    except ParsimoniousParseError:
        # we want to raise the nice parsimonious ParseError with all the explanation,
        # but we also want to raise something that isinstance(x, re.error)...
        # so we defined a doubly-inheriting class ParseError and we convert the exception
        # to that, taking care to keep the traceback (as described in
        # http://www.ianbicking.org/blog/2007/09/re-raising-exceptions.html)
        _exc_class, exc, tb = sys.exc_info()
        exc = ParseError(exc.text, exc.pos, exc.expr)
        raise exc.with_traceback(tb)
    compiled = compiler.compile(ast)
    regex = asm.assemble(compiled, syntax=syntax)
    return regex


def _wrap(name) -> Callable[[str, int, AnyStr, str], original_re.Pattern]:
    func = getattr(original_re, name)

    @functools.wraps(func)
    def wrapper(pattern, string, flags=0, syntax="python"):
        return func(re(pattern, syntax=syntax), string, flags=flags)

    return wrapper


def _wrap_sub(name) -> Callable[[str, str, int, int, str], original_re.Pattern]:
    func = getattr(original_re, name)

    @functools.wraps(func)
    def wrapper(pattern, repl, string, count=0, flags=0, syntax="python"):
        return func(
            re(pattern, syntax=syntax),
            repl=repl,
            string=string,
            count=count,
            flags=flags,
        )

    return wrapper


def compile(pattern, flags=0, syntax="python"):
    return original_re.compile(re(pattern, syntax=syntax), flags=flags)


search = _wrap("search")
match = _wrap("match")
fullmatch = _wrap("fullmatch")


def split(pattern, string, maxsplit=0, flags=0, syntax="python"):
    return original_re.split(
        re(pattern, syntax=syntax), string=string, maxsplit=maxsplit, flags=flags
    )


findall = _wrap("findall")
finditer = _wrap("finditer")
sub = _wrap_sub("sub")
subn = _wrap_sub("subn")

ESCAPE_RE = compile("['[' | ']']")


def escape(pattern):
    return ESCAPE_RE.sub(r"['\0']", pattern)


# TODO: when we get a cache of our own, clear it too
purge = original_re.purge

parser = argparse.ArgumentParser(
    description="Convert legacy regexp to kleenexp.",
    epilog="""usage: echo "Trololo lolo" | grep -P `ke "[#sl]Tro[0+ #space | 'lo']lo[#el]"`""",
)
parser.add_argument(
    "pattern",
    type=str,
    default="",
    help="a legacy regular expression (remember to escape it correctly)",
)
parser.add_argument(
    "--js",
    dest="syntax",
    action="store_const",
    const="javascript",
    default="python",
    help="output javascript regex syntax",
)


def main():
    args = parser.parse_args()
    try:
        print(re(args.pattern, syntax=args.syntax), end="")
        return 0
    except error:
        t, v, _tb = sys.exc_info()
        print("".join(traceback.format_exception_only(t, v)).strip(), file=sys.stderr)
        return 1


if __name__ == "__main__":
    sys.exit(main())



import pytest
from ke.asm import (
    assemble,
    Literal,
    Multiple,
    Either,
    Concat,
    CharacterClass,
    DIGIT,
    Capture,
    Setting,
    Boundary,
    START_LINE,
    START_STRING,
    WORD_BOUNDARY,
)


def test_literal():
    assert assemble(Literal("abc")) == "abc"
    assert assemble(Literal("^[a](b)$")) == r"\^\[a\]\(b\)\$"


def test_multiple():
    assert assemble(Multiple(0, 1, True, Literal("a"))) == "a?"
    assert assemble(Multiple(0, None, True, Literal("a"))) == "a*"
    assert assemble(Multiple(1, None, True, Literal("a"))) == "a+"
    assert assemble(Multiple(2, 2, True, Literal("a"))) == "a{2}"
    assert assemble(Multiple(None, 2, True, Literal("a"))) == "a{,2}"
    assert assemble(Multiple(2, None, True, Literal("a"))) == "a{2,}"
    assert assemble(Multiple(2, 5, True, Literal("a"))) == "a{2,5}"


def test_multiple_subexpression():
    assert assemble(Multiple(0, 1, True, Literal("abc"))) == "(?:abc)?"
    assert assemble(Multiple(0, None, True, Literal("abc"))) == "(?:abc)*"
    assert assemble(Multiple(1, None, True, Literal("abc"))) == "(?:abc)+"
    assert (
        assemble(Multiple(0, 1, True, Multiple(2, 3, True, Literal("a"))))
        == "(?:a{2,3})?"
    )


def test_multiple_nongreedy():
    assert assemble(Multiple(0, 1, False, Literal("a"))) == "a??"
    assert assemble(Multiple(0, None, False, Literal("a"))) == "a*?"
    assert assemble(Multiple(1, None, False, Literal("a"))) == "a+?"
    assert assemble(Multiple(2, 2, False, Literal("a"))) == "a{2}?"


def test_either():
    assert assemble(Either(list(map(Literal, "abc")))) == "a|b|c"
    assert assemble(Either(list(map(Literal, ["123", "45", ""])))) == "123|45|"
    assert (
        assemble(Multiple(1, None, True, Either(list(map(Literal, ["123", "45", ""])))))
        == "(?:123|45|)+"
    )
    assert (
        assemble(Concat([Literal("a"), Either([Literal("b"), Literal("c")])]))
        == "a(?:b|c)"
    )
    assert assemble(Concat([Either([Literal("b"), Literal("c")])])) == "b|c"


def test_concat():
    assert assemble(Concat(list(map(Literal, "abc")))) == "abc"
    assert assemble(Concat(list(map(Literal, ["123", "45", ""])))) == "12345"
    assert (
        assemble(Concat([Literal("123"), Multiple(0, 1, True, Literal("abc"))]))
        == "123(?:abc)?"
    )


def test_character_class():
    assert assemble(DIGIT) == r"\d"
    assert assemble(Concat([DIGIT, DIGIT])) == r"\d\d"
    assert assemble(Concat([DIGIT, Multiple(0, None, True, DIGIT)])) == r"\d\d*"
    assert assemble(CharacterClass(["a"], False)) == r"a"
    assert assemble(CharacterClass(["\n"], False)) == r"\n"
    assert assemble(CharacterClass(["["], False)) == r"\["
    assert assemble(CharacterClass(["{"], False)) == r"\{"
    assert assemble(CharacterClass(["}"], False)) == r"\}"
    assert assemble(CharacterClass(["?"], False)) == r"\?"
    assert assemble(CharacterClass(["a", "b"], False)) == r"[ab]"
    assert assemble(CharacterClass(["b", "a"], False)) == r"[ab]"
    assert (
        assemble(CharacterClass(["^", "\n", "\t", "-", "[", "]", "\\"], False))
        == r"[\-\[\\\]\^\n\t]"
    )
    assert assemble(CharacterClass(["a"], inverted=True)) == r"[^a]"
    assert assemble(CharacterClass(["a", "b"], inverted=True)) == r"[^ab]"
    assert assemble(CharacterClass([["a", "z"]], False)) == r"[a-z]"
    assert assemble(CharacterClass([["a", "z"], ["0", "5"]], False)) == r"[0-5a-z]"
    assert (
        assemble(CharacterClass([["a", "z"], ["0", "5"], "X"], False)) == r"[0-5Xa-z]"
    )
    assert assemble(CharacterClass([["a", "z"]], inverted=True)) == r"[^a-z]"
    assert (
        assemble(CharacterClass([["a", "z"], ["0", "5"]], inverted=True))
        == r"[^0-5a-z]"
    )


def test_capture():
    assert assemble(Capture(None, DIGIT)) == r"(\d)"
    assert (
        assemble(
            Concat([Literal("No. "), Capture("number", Multiple(1, None, True, DIGIT))])
        )
        == r"No\. (?P<number>\d+)"
    )


def test_setting():
    assert assemble(Setting("m", Literal("a"))) == "(?m)a"
    assert assemble(Multiple(0, 1, True, Setting("m", Literal("ab")))) == "(?m)(?:ab)?"


def test_boundary():
    assert assemble(Boundary(r"\b", r"\B")) == r"\b"
    assert Boundary(r"\b", r"\B").invert() == Boundary(r"\B", r"\b")
    with pytest.raises(ValueError):
        Boundary(r"\A", None).invert()
    assert assemble(START_LINE) == r"^"
    assert assemble(START_STRING) == r"\A"
    assert assemble(WORD_BOUNDARY) == r"\b"



import re

from ke.parser import (
    Parser,
    Concat,
    Either,
    Def,
    Operator,
    Macro,
    Range,
    Literal,
    Nothing,
)
from ke import asm
from ke.errors import CompileError

parser = Parser()

EMPTY = asm.Literal("")
EMPTY_CONCAT = asm.Concat([])

builtin_macros = {
    "#any": asm.ANY,
    "#newline_character": asm.NEWLINE,
    "#newline": asm.Either([asm.NEWLINE, asm.Literal("\r\n")]),
    "#not_newline": asm.NEWLINE.invert(),
    "#any_at_all": asm.Either([asm.ANY, asm.NEWLINE]),
    "#linefeed": asm.LINEFEED,
    "#carriage_return": asm.CARRIAGE_RETURN,
    "#windows_newline": asm.Literal("\r\n"),
    "#tab": asm.TAB,
    "#digit": asm.DIGIT,
    "#letter": asm.LETTER,
    "#lowercase": asm.LOWERCASE,
    "#uppercase": asm.UPPERCASE,
    "#space": asm.SPACE,
    "#token_character": asm.TOKEN_CHARACTER,
    "#start_string": asm.START_STRING,
    "#end_string": asm.END_STRING,
    "#start_line": asm.START_LINE,
    "#end_line": asm.END_LINE,
    "#word_boundary": asm.WORD_BOUNDARY,
    "#quote": asm.Literal("'"),
    "#double_quote": asm.Literal('"'),
    "#left_brace": asm.Literal("["),
    "#right_brace": asm.Literal("]"),
    "#vertical_tab": asm.Literal("\v"),
    "#formfeed": asm.Literal("\f"),
    "#bell": asm.Literal("\a"),
    "#backspace": asm.Literal("\b"),
}
for (
    name
) in "linefeed carriage_return tab digit letter lowercase uppercase space token_character word_boundary".split():
    macro = builtin_macros["#" + name]
    builtin_macros["#not_" + name] = macro.invert()
for names in """\
linefeed lf
carriage_return cr
tab t
digit d
letter l
lowercase lc
uppercase uc
space s
token_character tc
word_boundary wb""".splitlines():
    long, short = names.split()
    builtin_macros["#" + short] = builtin_macros["#" + long]
    builtin_macros["#n" + short] = builtin_macros["#not_" + long]
for names in """\
any a
any_at_all aaa
newline n
newline_character nc
not_newline nn
windows_newline crlf
start_string ss
end_string es
start_line sl
end_line el
quote q
double_quote dq
left_brace lb
right_brace rb""".splitlines():
    long, short = names.split()
    builtin_macros["#" + short] = builtin_macros["#" + long]


def invert_operator(n, expr):
    if n is not None:
        raise CompileError("Invert operator does not accept name")
    if isinstance(expr, asm.Literal) and len(expr.string) == 1:
        return asm.CharacterClass([expr.string], True)

    try:
        return expr.invert()
    except AttributeError:
        raise CompileError(
            "Expression %s cannot be inverted (maybe try [not lookahead <expression>]?)"
            % expr.to_regex(
                syntax="python"
            )  # TODO: maybe pass syntax here for better message?
        )


builtin_operators = {
    "capture": asm.Capture,
    "not": invert_operator,
    "lookahead": asm.Lookahead,
    "lookbehind": asm.Lookbehind,
}
for names in """\
capture c
not n
lookahead la""".splitlines():
    long, short = names.split()
    builtin_operators[short] = builtin_operators[long]


def compile(ast):
    macros = dict(builtin_macros)
    return asm.Setting("", compile_ast(ast, macros))


def compile_ast(ast, macros):
    return converters[type(ast)](ast, macros)


def compile_concat(concat, macros):
    defs = [x for x in concat.items if isinstance(x, Def)]
    regexes = [x for x in concat.items if not isinstance(x, Def)]

    for d in defs:
        if d.name in macros:
            raise KeyError("Macro %s already defined" % d.name)
        macros[d.name] = compile_ast(d.subregex, macros)
    compiled = [compile_ast(s, macros) for s in regexes]
    compiled = [x for x in compiled if is_not_empty(x)]
    for d in defs:
        del macros[d.name]

    if not compiled:
        return EMPTY
    if len(compiled) == 1:
        compiled, = compiled
        return compiled
    return asm.Concat(compiled)


def is_not_empty(node):
    return node != EMPTY and node != EMPTY_CONCAT


def def_error(d, macros):
    # TODO: AST transformation that takes Defs out of Either, etc' so we can define them anywhere with sane semantics
    raise ValueError(
        "temporarily, macro definition is only allowed directly under Concat([])"
    )


def compile_either(e, macros):
    compiled = [compile_ast(s, macros) for s in e.items]
    if all(is_single_char(c) for c in compiled):
        characters = []
        for c in compiled:
            if isinstance(c, asm.Literal) and len(c.string) == 1:
                characters.append(c.string)
            elif isinstance(c, asm.CharacterClass):
                characters += c.characters
        return asm.CharacterClass(characters, False)
    return asm.Either(compiled)


def is_single_char(c):
    return (isinstance(c, asm.Literal) and len(c.string) == 1) or isinstance(
        c, asm.CharacterClass
    )


REPEAT_OPERATOR = re.compile(r"(?:(\d+)-(\d+)|(\d+)\+|(\d+))$")


def compile_operator(o, macros):
    if o.op_name == "comment":
        return EMPTY
    sub = compile_ast(o.subregex, macros)
    if not is_not_empty(sub):
        raise CompileError("Operator %s not allowed to have empty body" % o.op_name)
    m = REPEAT_OPERATOR.match(o.op_name)
    if m:
        min, max, min2, exact = m.groups()
        if min2:
            min = int(min2)
            max = None
        elif exact:
            min = max = int(exact)
        else:
            min = int(min)
            max = int(max)
        if min == max == 0:
            return EMPTY
        return asm.Multiple(min, max, True, sub)
    if o.op_name not in builtin_operators:
        raise CompileError("Operator %s does not exist" % o.op_name)
    return builtin_operators[o.op_name](o.name, sub)


def compile_macro(macro, macros):
    if macro.name not in macros:
        raise CompileError(
            "Macro %s does not exist, perhaps you defined it in the wrong scope?"
            % macro.name
        )
    return macros[macro.name]


def compile_range(range, _):
    if character_category(range.start) != character_category(range.end):
        raise CompileError(
            "Range start and end not of the same category: '%s' is a %s but '%s' is a %s"
            % (
                range.start,
                character_category(range.start),
                range.end,
                character_category(range.end),
            )
        )
    if range.start >= range.end:
        raise CompileError(
            "Range start not before range end: '%s' >= '%s'" % (range.start, range.end)
        )
    return asm.CharacterClass([(range.start, range.end)], False)


def character_category(char):
    if char.islower():
        return "LOWERCASE"
    if char.isupper():
        return "UPPERCASE"
    if char.isdigit():
        return "DIGIT"
    assert False, "Unknown character class: %s" % char


converters = {
    Concat: compile_concat,
    Either: compile_either,
    Def: def_error,
    Operator: compile_operator,
    Macro: compile_macro,
    Range: compile_range,
    Literal: lambda l, _: asm.Literal(l.string),
    Nothing: lambda _n, _: EMPTY,
}


def add_builtin_macro(long, short, definition):
    ast = compile_ast(parser.parse(definition), builtin_macros)
    builtin_macros[long] = ast
    if short:
        builtin_macros[short] = ast


add_builtin_macro("#integer", "#int", "[[0-1 '-'] [1+ #digit]]")
add_builtin_macro("#unsigned_integer", "#uint", "[1+ #digit]")
add_builtin_macro("#real", None, "[#int [0-1 '.' #uint]]")
add_builtin_macro(
    "#float",
    None,
    "[[0-1 '-'] [[#uint '.' [0-1 #uint] | '.' #uint] [0-1 #exponent] | #int #exponent] #exponent=[['e' | 'E'] [0-1 ['+' | '-']] #uint]]",
)
add_builtin_macro("#hex_digit", "#hexd", "[#digit | #a..f | #A..F]")
add_builtin_macro("#hex_number", "#hexn", "[1+ #hex_digit]")
# this is not called #word because in legacy regex \w (prounounced "word character") is #token_character and I fear confusion
add_builtin_macro("#letters", None, "[1+ #letter]")
add_builtin_macro("#token", None, "[#letter | '_'][0+ #token_character]")
add_builtin_macro("#capture_0+_any", "#c0", "[capture 0+ #any]")
add_builtin_macro("#capture_1+_any", "#c1", "[capture 1+ #any]")
